@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2001-2015. All Rights Reserved.
+%% Copyright Ericsson AB 2001-2017. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -20,9 +20,7 @@
 -module(cover).
 
 %%
-%% This module implements the Erlang coverage tool. The module named
-%% cover_web implements a user interface for the coverage tool to run
-%% under webtool.
+%% This module implements the Erlang coverage tool.
 %% 
 %% ARCHITECTURE
 %% The coverage tool consists of one process on each node involved in
@@ -575,7 +573,7 @@ call(Request) ->
     Ref = erlang:monitor(process,?SERVER),
     receive {'DOWN', Ref, _Type, _Object, noproc} -> 
 	    erlang:demonitor(Ref),
-	    start(),
+            {ok,_} = start(),
 	    call(Request)
     after 0 ->
 	    ?SERVER ! {self(),Request},
@@ -591,7 +589,9 @@ call(Request) ->
     end.
 
 reply(From, Reply) ->
-    From ! {?SERVER,Reply}.
+    From ! {?SERVER,Reply},
+    ok.
+
 is_from(From) ->
     is_pid(From).
 
@@ -617,9 +617,11 @@ remote_call(Node,Request) ->
     end.
     
 remote_reply(Proc,Reply) when is_pid(Proc) ->
-    Proc ! {?SERVER,Reply};
+    Proc ! {?SERVER,Reply},
+    ok;
 remote_reply(MainNode,Reply) ->
-    {?SERVER,MainNode} ! {?SERVER,Reply}.
+    {?SERVER,MainNode} ! {?SERVER,Reply},
+    ok.
 
 %%%----------------------------------------------------------------------
 %%% cover_server on main node
@@ -629,14 +631,16 @@ init_main(Starter) ->
     register(?SERVER,self()),
     %% Having write concurrancy here gives a 40% performance boost
     %% when collect/1 is called. 
-    ets:new(?COVER_TABLE, [set, public, named_table
-			   ,{write_concurrency, true}
-			  ]),
-    ets:new(?COVER_CLAUSE_TABLE, [set, public, named_table]),
-    ets:new(?BINARY_TABLE, [set, public, named_table]),
-    ets:new(?COLLECTION_TABLE, [set, public, named_table]),
-    ets:new(?COLLECTION_CLAUSE_TABLE, [set, public, named_table]),
-    net_kernel:monitor_nodes(true),
+    ?COVER_TABLE = ets:new(?COVER_TABLE, [set, public, named_table,
+                                          {write_concurrency, true}]),
+    ?COVER_CLAUSE_TABLE = ets:new(?COVER_CLAUSE_TABLE, [set, public,
+                                                        named_table]),
+    ?BINARY_TABLE = ets:new(?BINARY_TABLE, [set, public, named_table]),
+    ?COLLECTION_TABLE = ets:new(?COLLECTION_TABLE, [set, public,
+                                                    named_table]),
+    ?COLLECTION_CLAUSE_TABLE = ets:new(?COLLECTION_CLAUSE_TABLE, [set, public,
+                                                                  named_table]),
+    ok = net_kernel:monitor_nodes(true),
     Starter ! {?SERVER,started},
     main_process_loop(#main_state{}).
 
@@ -674,7 +678,7 @@ main_process_loop(State) ->
 		    Imported = do_import_to_table(Fd,File,
 						  State#main_state.imported),
 		    reply(From, ok),
-		    file:close(Fd),
+		    ok = file:close(Fd),
 		    main_process_loop(State#main_state{imported=Imported});
 		{error,Reason} ->
 		    reply(From, {error, {cant_open_file,File,Reason}}),
@@ -872,11 +876,12 @@ main_process_loop(State) ->
 
 init_remote(Starter,MainNode) ->
     register(?SERVER,self()),
-    ets:new(?COVER_TABLE, [set, public, named_table
-			   %% write_concurrency here makes otp_8270 break :(
-			   %,{write_concurrency, true}
-			  ]),
-    ets:new(?COVER_CLAUSE_TABLE, [set, public, named_table]),
+    %% write_concurrency here makes otp_8270 break :(
+    ?COVER_TABLE = ets:new(?COVER_TABLE, [set, public, named_table
+                                          %,{write_concurrency, true}
+                                         ]),
+    ?COVER_CLAUSE_TABLE = ets:new(?COVER_CLAUSE_TABLE, [set, public,
+                                                        named_table]),
     Starter ! {self(),started},
     remote_process_loop(#remote_state{main_node=MainNode}).
 
@@ -909,11 +914,11 @@ remote_process_loop(State) ->
 			  '_' -> [M || {M,_} <- State#remote_state.compiled];
 			  _ -> Modules0
 		      end,
-	   spawn(fun() ->
-			 ?SPAWN_DBG(remote_collect, 
-				    {Modules, CollectorPid, From}),
-			 do_collect(Modules, CollectorPid, From)
-		 end),
+            spawn(fun() ->
+                          ?SPAWN_DBG(remote_collect, 
+                                     {Modules, CollectorPid, From}),
+                          do_collect(Modules, CollectorPid, From)
+                  end),
 	    remote_process_loop(State);
 
 	{remote,stop} ->
@@ -954,13 +959,13 @@ remote_process_loop(State) ->
     end.
 
 do_collect(Modules, CollectorPid, From) ->
-    pmap(
-      fun(Module) ->
-	      Pattern = {#bump{module=Module, _='_'}, '$1'},
-	      MatchSpec = [{Pattern,[{'=/=','$1',0}],['$_']}],
-	      Match = ets:select(?COVER_TABLE,MatchSpec,?CHUNK_SIZE),
-	      send_chunks(Match, CollectorPid, [])
-      end,Modules),
+    _ = pmap(
+          fun(Module) ->
+                  Pattern = {#bump{module=Module, _='_'}, '$1'},
+                  MatchSpec = [{Pattern,[{'=/=','$1',0}],['$_']}],
+                  Match = ets:select(?COVER_TABLE,MatchSpec,?CHUNK_SIZE),
+                  send_chunks(Match, CollectorPid, [])
+          end,Modules),
     CollectorPid ! done,
     remote_reply(From, ok).
 
@@ -996,20 +1001,20 @@ get_downs(Mons) ->
     end.
 
 reload_originals(Compiled) ->
-    Modules = [M || {M,_} <- Compiled],
-    pmap(fun do_reload_original/1, Modules).
+    _ = pmap(fun do_reload_original/1, [M || {M,_} <- Compiled]),
+    ok.
 
 do_reload_original(Module) ->
     case code:which(Module) of
 	?TAG ->
-	    code:purge(Module),     % remove code marked as 'old'
-	    code:delete(Module),    % mark cover compiled code as 'old'
+	    _ = code:purge(Module),     % remove code marked as 'old'
+	    _ = code:delete(Module),    % mark cover compiled code as 'old'
 	    %% Note: original beam code must be loaded before the cover
 	    %% compiled code is purged, in order to for references to
 	    %% 'fun M:F/A' and %% 'fun F/A' funs to be correct (they
 	    %% refer to (M:)F/A in the *latest* version  of the module)
-	    code:load_file(Module), % load original code
-	    code:purge(Module);     % remove cover compiled code
+            _ = code:load_file(Module), % load original code
+	    _ = code:purge(Module);     % remove cover compiled code
 	_ ->
 	    ignore
     end.
@@ -1221,12 +1226,13 @@ remote_reset(Module,Nodes) ->
 
 %% Collect data from remote nodes - used for analyse or stop(Node)
 remote_collect(Modules,Nodes,Stop) ->
-    pmap(fun(Node) -> 
-		 ?SPAWN_DBG(remote_collect, 
-			    {Modules, Nodes, Stop}),
-		 do_collection(Node, Modules, Stop)
-	 end,
-	 Nodes).
+    _ = pmap(
+          fun(Node) -> 
+                  ?SPAWN_DBG(remote_collect, 
+                             {Modules, Nodes, Stop}),
+                  do_collection(Node, Modules, Stop)
+          end, Nodes),
+    ok.
 
 do_collection(Node, Module, Stop) ->
     CollectorPid = spawn(fun collector_proc/0),
@@ -1262,8 +1268,8 @@ insert_in_collection_table([]) ->
 insert_in_collection_table(Key,Val) ->
     case ets:member(?COLLECTION_TABLE,Key) of
 	true ->
-	    ets:update_counter(?COLLECTION_TABLE,
-			       Key,Val);
+	    _ = ets:update_counter(?COLLECTION_TABLE, Key,Val),
+            ok;
 	false ->
 	    %% Make sure that there are no race conditions from ets:member
 	    case ets:insert_new(?COLLECTION_TABLE,{Key,Val}) of
@@ -1474,7 +1480,7 @@ do_compile_beam(Module,BeamFile0,State) ->
 		    {ok,Module,BeamFile};
 		error ->
 		    {error, BeamFile};
-		{error,Reason} -> % no abstract code
+		{error,Reason} -> % no abstract code or no 'file' attribute
 		    {error, {Reason, BeamFile}}
 	    end;
 	{error,no_beam} ->
@@ -1537,32 +1543,11 @@ do_compile_beam1(Module,Beam,UserOptions) ->
 	    {error,E};
 	{raw_abstract_v1,Code} ->
             Forms0 = epp:interpret_file_attribute(Code),
-	    {Forms,Vars} = transform(Forms0, Module),
-
-	    %% We need to recover the source from the compilation
-	    %% info otherwise the newly compiled module will have
-	    %% source pointing to the current directory
-	    SourceInfo = get_source_info(Module, Beam),
-
-	    %% Compile and load the result
-	    %% It's necessary to check the result of loading since it may
-	    %% fail, for example if Module resides in a sticky directory
-	    {ok, Module, Binary} = compile:forms(Forms, SourceInfo ++ UserOptions),
-	    case code:load_binary(Module, ?TAG, Binary) of
-		{module, Module} ->
-		    
-		    %% Store info about all function clauses in database
-		    InitInfo = lists:reverse(Vars#vars.init_info),
-		    ets:insert(?COVER_CLAUSE_TABLE, {Module, InitInfo}),
-		    
-		    %% Store binary code so it can be loaded on remote nodes
-		    ets:insert(?BINARY_TABLE, {Module, Binary}),
-
-		    {ok, Module};
-		
-		_Error ->
-		    do_clear(Module),
-		    error
+	    case find_main_filename(Forms0) of
+		{ok,MainFile} ->
+		    do_compile_beam2(Module,Beam,UserOptions,Forms0,MainFile);
+		Error ->
+		    Error
 	    end;
 	{_VSN,_Code} ->
 	    %% Wrong version of abstract code. Just report that there
@@ -1579,6 +1564,35 @@ get_abstract_code(Module, Beam) ->
 	Error -> Error
     end.
 
+do_compile_beam2(Module,Beam,UserOptions,Forms0,MainFile) ->
+    {Forms,Vars} = transform(Forms0, Module, MainFile),
+
+    %% We need to recover the source from the compilation
+    %% info otherwise the newly compiled module will have
+    %% source pointing to the current directory
+    SourceInfo = get_source_info(Module, Beam),
+
+    %% Compile and load the result
+    %% It's necessary to check the result of loading since it may
+    %% fail, for example if Module resides in a sticky directory
+    {ok, Module, Binary} = compile:forms(Forms, SourceInfo ++ UserOptions),
+    case code:load_binary(Module, ?TAG, Binary) of
+	{module, Module} ->
+
+	    %% Store info about all function clauses in database
+	    InitInfo = lists:reverse(Vars#vars.init_info),
+	    ets:insert(?COVER_CLAUSE_TABLE, {Module, InitInfo}),
+
+	    %% Store binary code so it can be loaded on remote nodes
+	    ets:insert(?BINARY_TABLE, {Module, Binary}),
+
+	    {ok, Module};
+
+	_Error ->
+	    do_clear(Module),
+	    error
+    end.
+
 get_source_info(Module, Beam) ->
     Compile = get_compile_info(Module, Beam),
     case lists:keyfind(source, 1, Compile) of
@@ -1601,8 +1615,7 @@ get_compile_info(Module, Beam) ->
 		[]
     end.
 
-transform(Code, Module) ->
-    MainFile=find_main_filename(Code),
+transform(Code, Module, MainFile) ->
     Vars0 = #vars{module=Module},
     {ok,MungedForms,Vars} = transform_2(Code,[],Vars0,MainFile,on),
     {MungedForms,Vars}.
@@ -1610,9 +1623,12 @@ transform(Code, Module) ->
 %% Helpfunction which returns the first found file-attribute, which can
 %% be interpreted as the name of the main erlang source file.
 find_main_filename([{attribute,_,file,{MainFile,_}}|_]) ->
-    MainFile;
+    {ok,MainFile};
 find_main_filename([_|Rest]) ->
-    find_main_filename(Rest).
+    find_main_filename(Rest);
+find_main_filename([]) ->
+    {error, no_file_attribute}.
+
 
 transform_2([Form0|Forms],MungedForms,Vars,MainFile,Switch) ->
     Form = expand(Form0),
@@ -1994,9 +2010,7 @@ munge_expr({lc,Line,Expr,Qs}, Vars) ->
     {MungedQs, Vars3} = munge_qualifiers(Qs, Vars2),
     {{lc,Line,MungedExpr,MungedQs}, Vars3};
 munge_expr({bc,Line,Expr,Qs}, Vars) ->
-    {bin,BLine,[{bin_element,EL,Val,Sz,TSL}|Es]} = Expr,
-    Expr2 = {bin,BLine,[{bin_element,EL,?BLOCK1(Val),Sz,TSL}|Es]},
-    {MungedExpr,Vars2} = munge_expr(Expr2, Vars),
+    {MungedExpr,Vars2} = munge_expr(?BLOCK1(Expr), Vars),
     {MungedQs, Vars3} = munge_qualifiers(Qs, Vars2),
     {{bc,Line,MungedExpr,MungedQs}, Vars3};
 munge_expr({block,Line,Body}, Vars) ->
@@ -2039,7 +2053,7 @@ munge_expr({bin_element,Line,Value,Size,TypeSpecifierList}, Vars) ->
     {MungedValue,Vars2} = munge_expr(Value, Vars),
     {MungedSize,Vars3} = munge_expr(Size, Vars2),
     {{bin_element,Line,MungedValue,MungedSize,TypeSpecifierList},Vars3};
-munge_expr(Form, Vars) -> % var|char|integer|float|string|atom|nil|eof|default
+munge_expr(Form, Vars) ->
     {Form, Vars}.
 
 munge_exprs([Expr|Exprs], Vars, MungedExprs) when Vars#vars.is_guard=:=true,
@@ -2400,8 +2414,8 @@ do_analyse_to_file1(Module, OutFile, ErlFile, HTML) ->
 	{ok, InFd} ->
 	    case file:open(OutFile, [write,raw,delayed_write]) of
 		{ok, OutFd} ->
+                    Enc = encoding(ErlFile),
 		    if HTML -> 
-                           Encoding = encoding(ErlFile),
                            Header =
                                ["<!DOCTYPE HTML PUBLIC "
                                 "\"-//W3C//DTD HTML 3.2 Final//EN\">\n"
@@ -2409,13 +2423,14 @@ do_analyse_to_file1(Module, OutFile, ErlFile, HTML) ->
                                 "<head>\n"
                                 "<meta http-equiv=\"Content-Type\""
                                 " content=\"text/html; charset=",
-                                Encoding,"\"/>\n"
+                                html_encoding(Enc),"\"/>\n"
                                 "<title>",OutFile,"</title>\n"
                                 "</head>"
                                 "<body style='background-color: white;"
                                 " color: black'>\n"
                                 "<pre>\n"],
-                           file:write(OutFd,Header);
+                            H1Bin = unicode:characters_to_binary(Header,Enc,Enc),
+                            ok = file:write(OutFd,H1Bin);
 		       true -> ok
 		    end,
 		    
@@ -2424,31 +2439,35 @@ do_analyse_to_file1(Module, OutFile, ErlFile, HTML) ->
                    Timestamp =
                        io_lib:format("~p-~s-~s at ~s:~s:~s",
                                      [Y,
-                                      string:right(integer_to_list(Mo), 2, $0),
-                                      string:right(integer_to_list(D),  2, $0),
-                                      string:right(integer_to_list(H),  2, $0),
-                                      string:right(integer_to_list(Mi), 2, $0),
-                                      string:right(integer_to_list(S),  2, $0)]),
-                    file:write(OutFd,
-                               ["File generated from ",ErlFile," by COVER ",
+                                      string:pad(integer_to_list(Mo), 2, leading, $0),
+                                      string:pad(integer_to_list(D),  2, leading, $0),
+                                      string:pad(integer_to_list(H),  2, leading, $0),
+                                      string:pad(integer_to_list(Mi), 2, leading, $0),
+                                      string:pad(integer_to_list(S),  2, leading, $0)]),
+
+                    H2Bin = unicode:characters_to_binary(
+                              ["File generated from ",ErlFile," by COVER ",
                                 Timestamp,"\n\n"
                                 "**************************************"
                                 "**************************************"
-                                "\n\n"]),
+                                "\n\n"],
+                              Enc, Enc),
+                    ok = file:write(OutFd, H2Bin),
 
 		    Pattern = {#bump{module=Module,line='$1',_='_'},'$2'},
 		    MS = [{Pattern,[{is_integer,'$1'},{'>','$1',0}],[{{'$1','$2'}}]}],
-		    CovLines = lists:keysort(1,ets:select(?COLLECTION_TABLE, MS)),
+                    CovLines0 =
+                        lists:keysort(1, ets:select(?COLLECTION_TABLE, MS)),
+                    CovLines = merge_dup_lines(CovLines0),
 		    print_lines(Module, CovLines, InFd, OutFd, 1, HTML),
 		    
-		    if
-			HTML ->
-			    file:write(OutFd, "</pre>\n</body>\n</html>\n");
+		    if HTML ->
+                           ok = file:write(OutFd, "</pre>\n</body>\n</html>\n");
 		       true -> ok
 		    end,
 
-		    file:close(OutFd),
-		    file:close(InFd),
+		    ok = file:close(OutFd),
+		    ok = file:close(InFd),
 
 		    {ok, OutFile};
 
@@ -2460,40 +2479,43 @@ do_analyse_to_file1(Module, OutFile, ErlFile, HTML) ->
 	    {error, {file, ErlFile, Reason}}
     end.
 
+merge_dup_lines(CovLines) ->
+    merge_dup_lines(CovLines, []).
+merge_dup_lines([{L, N}|T], [{L, NAcc}|TAcc]) ->
+    merge_dup_lines(T, [{L, NAcc + N}|TAcc]);
+merge_dup_lines([{L, N}|T], Acc) ->
+    merge_dup_lines(T, [{L, N}|Acc]);
+merge_dup_lines([], Acc) ->
+    lists:reverse(Acc).
 
 print_lines(Module, CovLines, InFd, OutFd, L, HTML) ->
     case file:read_line(InFd) of
 	eof ->
 	    ignore;
-	{ok,"%"++_=Line} ->		 %Comment line - not executed.
-	    file:write(OutFd, [tab(),escape_lt_and_gt(Line, HTML)]),
-	    print_lines(Module, CovLines, InFd, OutFd, L+1, HTML);
 	{ok,RawLine} ->
 	    Line = escape_lt_and_gt(RawLine,HTML),
 	    case CovLines of
 	       [{L,N}|CovLines1] ->
-		    %% N = lists:foldl(fun([Ni], Nacc) -> Nacc+Ni end, 0, Ns),
-		    if
-			N=:=0, HTML=:=true ->
-			    LineNoNL = Line -- "\n",
-			    Str = "     0",
-			    %%Str = string:right("0", 6, 32),
-			    RedLine = ["<font color=red>",Str,fill1(),
-				       LineNoNL,"</font>\n"],
-			    file:write(OutFd, RedLine);
-			N<1000000 ->
-			    Str = string:right(integer_to_list(N), 6, 32),
-			    file:write(OutFd, [Str,fill1(),Line]);
-			N<10000000 ->
-			    Str = integer_to_list(N),
-			    file:write(OutFd, [Str,fill2(),Line]);
-			true ->
-			    Str = integer_to_list(N),
-			    file:write(OutFd, [Str,fill3(),Line])
-		    end,
+                    if N=:=0, HTML=:=true ->
+                           LineNoNL = Line -- "\n",
+                           Str = "     0",
+                           %%Str = string:pad("0", 6, leading, $\s),
+                           RedLine = ["<font color=red>",Str,fill1(),
+                                      LineNoNL,"</font>\n"],
+                           ok = file:write(OutFd, RedLine);
+                       N < 1000000 ->
+                           Str = string:pad(integer_to_list(N), 6, leading, $\s),
+                           ok = file:write(OutFd, [Str,fill1(),Line]);
+                       N < 10000000 ->
+                           Str = integer_to_list(N),
+                           ok = file:write(OutFd, [Str,fill2(),Line]);
+                       true ->
+                           Str = integer_to_list(N),
+                           ok = file:write(OutFd, [Str,fill3(),Line])
+                    end,
 		    print_lines(Module, CovLines1, InFd, OutFd, L+1, HTML);
-		_ ->
-		    file:write(OutFd, [tab(),Line]),
+		_ ->                            %Including comment lines
+		    ok = file:write(OutFd, [tab(),Line]),
 		    print_lines(Module, CovLines, InFd, OutFd, L+1, HTML)
 	    end
     end.
@@ -2533,7 +2555,7 @@ do_export(Module, OutFile, From, State) ->
 				{error,{not_cover_compiled,Module}}
 			end
 		end,
-	    file:close(Fd),
+	    ok = file:close(Fd),
 	    reply(From, Reply);
 	{error,Reason} ->
 	    reply(From, {error, {cant_open_file,OutFile,Reason}})
@@ -2575,10 +2597,9 @@ write(Element,Fd) ->
     case byte_size(Bin) of
 	Size when Size > 255 ->
 	    SizeBin = term_to_binary({'$size',Size}),
-	    file:write(Fd,
-                       <<(byte_size(SizeBin)):8,SizeBin/binary,Bin/binary>>);
+	    ok = file:write(Fd, <<(byte_size(SizeBin)):8,SizeBin/binary,Bin/binary>>);
 	Size ->
-	    file:write(Fd,<<Size:8,Bin/binary>>)
+	    ok = file:write(Fd,<<Size:8,Bin/binary>>)
     end,
     ok.    
 
@@ -2741,16 +2762,22 @@ pmap_collect(Mons,Acc) ->
     end.
 
 %%%-----------------------------------------------------------------
-%%% Read encoding from source file
+%%% Decide which encoding to use when analyzing to file.
+%%% The target file contains the file path, so if either the file name
+%%% encoding or the encoding of the source file is utf8, then we need
+%%% to use utf8.
 encoding(File) ->
-    Encoding =
-       case epp:read_encoding(File) of
-           none ->
-               epp:default_encoding();
-           E ->
-               E
-       end,
-    html_encoding(Encoding).
+    case file:native_name_encoding() of
+        latin1 ->
+            case epp:read_encoding(File) of
+                none ->
+                    epp:default_encoding();
+                E ->
+                    E
+            end;
+        utf8 ->
+            utf8
+    end.
 
 html_encoding(latin1) ->
     "iso-8859-1";
